/**
 * File: search.cpp
 * ------------------
 * Defines a test harness for an interpolation search routine.
 */

#include <iostream>

#include "console.h"
#include "random.h"
#include "vector.h"
#include "set.h"
#include "strlib.h"
#include "error.h"
using namespace std;

static int search(Vector<double>& values, double key) {
    // replace this implementation with your own...
    return -1;
}

/**
 * This exercises basic functionality and confirms that search finds an 
 * element when it's supposed to, and fails to find an element when it's supposed to as well.
 * This particular test exercises search when the sorted vector contains a steady,
 * even distribution of values.
 *
 * @return void
 */
static const int kSampleSize = 5000000;
static void testEvenDistribution() {
	cout << "Begin TestEvenDistribution!" << endl;
	cout << "---------------------------" << endl;
	cout << "Allocating Vector<double>";
	Vector<double> numbers;
	cout << " and adding all numbers in the range [0, " << kSampleSize << "]." << endl;
	for (int i = 0; i <= kSampleSize; i++) {
		numbers += double(i);
	}
	
	try {
		cout << "Confirming that search finds numbers it's supposed to..." << endl;
		for (int j = 0; j <= kSampleSize; j++) {
			if (search(numbers, double(j)) != j) {
				error("Failed to find " + realToString(double(j)) + " in a vector known "
					  "to contain it.");
			}
		}
		cout << "    That seems to work." << endl;
		cout << "Confirming that search returns -1 when it's supposed to..." << endl;
		for (int i = 0; i <= kSampleSize; i++) {
			double value = i + 0.5;
			if (search(numbers, value) != -1) {
				error("We found the number " + realToString(value) +
					  " in a vector that doesn't actually included it");
			}
		}
		cout << "    That seems to work, too." << endl;
		cout << "TestEvenDistribution passes just fine.  Congrats!" << endl;
	} catch (ErrorException& e) {
		cout << e.getMessage() << endl;
	}

	cout << endl;
}

/**
 * Operates much like TestEvenDistribution does, except there's no guarantee 
 * that the numbers in the sorted vector are evenly distributed at all, because
 * all of its numbers are randomly generated.
 * 
 * @return void
 */

static const double kMinReal = 0.0;
static const double kMaxReal = 1000000000.0;
static void testRandomDistribution() {
	cout << "Begin TestRandomDistribution!" << endl;
	cout << "-----------------------------" << endl;
	cout << "Allocating Vector<double>";
	Vector<double> numbers;
	Set<double> knownNumbers;

	cout << " and appending " << kSampleSize << " random numbers to it." << endl;
	cout << "    [This takes a while, so be patient.]" << endl;
	for (int k = 0; k < kSampleSize; k++) {
		knownNumbers += randomReal(kMinReal, kMaxReal);
	}

	foreach (double n in knownNumbers) {
		numbers += n;
	}
	
	try {
		cout << "Confirming all of the numbers we added are actually present." << endl;
		foreach (double n in knownNumbers) {
			if (search(numbers, n) == -1)
				error("Vector should contain " + realToString(n) + ", but somehow " +
					  "search didn't find it.");
		}
		cout << "   and they are!" << endl;
		cout << "TestRandomDistribution passes just fine.  Congrats!" << endl;
	} catch (ErrorException& e) {
		cout << e.getMessage() << endl;
	}

	cout << endl;
}

static const double kOnlyNumberPresent = 55.55;
static void testAllSameDistribution() {
	cout << "Begin TestAllSameDistribution!" << endl;
	cout << "------------------------------" << endl;
	cout << "Allocating Vector<double>";
	Vector<double> values;
    cout << " and appending " << kSampleSize << " copies of the same number!" << endl;
	for (int i = 0; i < kSampleSize; i++) {
		values += kOnlyNumberPresent;
	}
		
	try {
		if (search(values, kOnlyNumberPresent) == -1)
			error("Can't find " + realToString(kOnlyNumberPresent) +
				  " in a vector that contains nothing but!");
		cout << "Have no problem finding the number (" << kOnlyNumberPresent << ") that's present." << endl;
		if (search(values, kOnlyNumberPresent + 0.1) != -1)
			error("Somehow finding a number that isn't present.");
		cout << "Confirmed that nothing larger than " << kOnlyNumberPresent << " is present." << endl;
		if (search(values, kOnlyNumberPresent - 0.1) != -1)
			error("Somehow finding a number that isn't present.");
		cout << "Confirmed that nothing smaller than " << kOnlyNumberPresent << " is present." << endl;
		cout << "TestAllSameDistribution passes with flying colors.  Congrats!" << endl;
	} catch (ErrorException& e) {
		cout << e.getMessage() << endl;
	}

	cout << endl;
}

/**
 * Unit test in place to aggressively exercise search to
 * confirm that it works.  We test the following:
 *
 * + We assert that search finds an element when it's known to exist.
 * + We assert that search doesn't find an element when it's known to be absent.
 * + We again assert that search finds an element when it's known to 
 *     be present, even though the distribution of values in the vector is imbalanced.
 * + We again assert that search doesn't find an element when it's known
 *     to be absent and the distribution of values in the vector is very imbalanced.
 *
 * @return int 0.
 */

int main() {
	testEvenDistribution();
	testRandomDistribution();
	testAllSameDistribution();
	return 0;
}

